class Constants:
    MODEL_FOLDER: str = "model"
    MODEL_CLASS_FILE: str = "model_class"
    MODEL_PARAM_FILE: str = "model_inputs_outputs.json"
    PIPELINE_FILE: str = "pipeline.pkl"
    INPUT_TYPE_CHECKER_FILE: str = "input_type_checker.pkl"
    OUTPUT_TYPE_CHECKER_FILE: str = "output_type_checker.pkl"
    EXTRA_FILES: str = "extra_files.json"
    SERIALISER_FILE: str = "serialiser.pkl"

    PYTHON_VERSION_FILE: str = "python_version.json"
    REQUIREMENTS_FILE: str = "requirements.txt"
